@@ -9,19 +9,26 @@
     :license: BSD, see LICENSE for more details.
 """
 from __future__ import with_statement, absolute_import
+import re
 import sys
 import time
 import sqlalchemy
 from math import ceil
-from types import MethodType
+from functools import partial
 from flask import _request_ctx_stack, abort
+from flask.signals import Namespace
 from operator import itemgetter
 from threading import Lock
 from sqlalchemy import orm
 from sqlalchemy.orm.exc import UnmappedClassError
+from sqlalchemy.orm.mapper import Mapper
+from sqlalchemy.orm.session import Session
+from sqlalchemy.orm.interfaces import MapperExtension, SessionExtension, \
+     EXT_CONTINUE
 from sqlalchemy.interfaces import ConnectionProxy
 from sqlalchemy.engine.url import make_url
 from sqlalchemy.ext.declarative import declarative_base
+from sqlalchemy.util import to_list
 
 # the best timer function for the platform
 if sys.platform == 'win32':
@@ -30,10 +37,27 @@ else:
     _timer = time.time
 
 
-def _create_scoped_session(db):
-    return orm.scoped_session(lambda: orm.create_session(autocommit=False,
-                                                         autoflush=False,
-                                                         bind=db.engine))
+_camelcase_re = re.compile(r'([A-Z]+)(?=[a-z0-9])')
+_signals = Namespace()
+
+
+models_committed = _signals.signal('models-committed')
+before_models_committed = _signals.signal('before-models-committed')
+
+
+def _create_scoped_session(db, options):
+    if options is None:
+        options = {}
+    return orm.scoped_session(partial(_SignallingSession, db, **options))
+
+
+def _make_table(db):
+    def _make_table(*args, **kwargs):
+        if len(args) > 1 and isinstance(args[1], db.Column):
+            args = (args[0], db.metadata) + args[2:]
+        return sqlalchemy.Table(*args, **kwargs)
+    return _make_table
+
 
 
 def _include_sqlalchemy(obj):
@@ -41,6 +65,7 @@ def _include_sqlalchemy(obj):
         for key in module.__all__:
             if not hasattr(obj, key):
                 setattr(obj, key, getattr(module, key))
+    obj.Table = _make_table(obj)
 
 
 class _DebugQueryTuple(tuple):
@@ -100,6 +125,62 @@ class _ConnectionDebugProxy(ConnectionProxy):
                     _calling_context(self.app_package))))
 
 
+class _SignalTrackingMapperExtension(MapperExtension):
+
+    def after_delete(self, mapper, connection, instance):
+        return self._record(mapper, instance, 'delete')
+
+    def after_insert(self, mapper, connection, instance):
+        return self._record(mapper, instance, 'insert')
+
+    def after_update(self, mapper, connection, instance):
+        return self._record(mapper, instance, 'update')
+
+    def _record(self, mapper, model, operation):
+        pk = tuple(mapper.primary_key_from_instance(model))
+        orm.object_session(model)._model_changes[pk] = (model, operation)
+        return EXT_CONTINUE
+
+
+class _SignalTrackingMapper(Mapper):
+
+    def __init__(self, *args, **kwargs):
+        extensions = to_list(kwargs.pop('extension', None), [])
+        extensions.append(_SignalTrackingMapperExtension())
+        kwargs['extension'] = extensions
+        Mapper.__init__(self, *args, **kwargs)
+
+
+class _SignallingSessionExtension(SessionExtension):
+
+    def before_commit(self, session):
+        d = session._model_changes
+        if d:
+            before_models_committed.send(session.app, changes=d.values())
+        return EXT_CONTINUE
+
+    def after_commit(self, session):
+        d = session._model_changes
+        if d:
+            models_committed.send(session.app, changes=d.values())
+            d.clear()
+        return EXT_CONTINUE
+
+    def after_rollback(self, session):
+        session._model_changes.clear()
+        return EXT_CONTINUE
+
+
+class _SignallingSession(Session):
+
+    def __init__(self, db, autocommit=False, autoflush=False, **options):
+        Session.__init__(self, autocommit=autocommit, autoflush=autoflush,
+                         extension=db.session_extensions,
+                         bind=db.engine, **options)
+        self.app = db.app or _request_ctx_stack.top.app
+        self._model_changes = {}
+
+
 def get_debug_queries():
     """In debug mode Flask-SQLAlchemy will log all the SQL queries sent
     to the database.  This information is available until the end of request
@@ -107,7 +188,7 @@ def get_debug_queries():
     one expected on errors or in unittesting.  If you don't want to enable
     the DEBUG mode for your unittests you can also enable the query
     recording by setting the ``'SQLALCHEMY_RECORD_QUERIES'`` config variable
-    to `True`.  This is automatically enablde if Flask is in testing mode.
+    to `True`.  This is automatically enabled if Flask is in testing mode.
 
     The value returned will be a list of named tuples with the following
     attributes:
@@ -279,8 +360,7 @@ class _QueryProperty(object):
     def __get__(self, obj, type):
         try:
             mapper = orm.class_mapper(type)
-            ctx = _request_ctx_stack.top
-            if mapper and ctx is not None:
+            if mapper:
                 return type.query_class(mapper, session=self.sa.session())
         except UnmappedClassError:
             return None
@@ -323,6 +403,21 @@ class _EngineConnector(object):
             return rv
 
 
+class _ModelTableNameDescriptor(object):
+
+    def __get__(self, obj, type):
+        tablename = type.__dict__.get('__tablename__')
+        if not tablename:
+            def _join(match):
+                word = match.group()
+                if len(word) > 1:
+                    return ('_%s_%s' % (word[:-1], word[-1])).lower()
+                return '_' + word.lower()
+            tablename = _camelcase_re.sub(_join, type.__name__).lstrip('_')
+            setattr(type, '__tablename__', tablename)
+        return tablename
+
+
 class Model(object):
     """Baseclass for custom user models."""
 
@@ -334,6 +429,11 @@ class Model(object):
     #: database for instances of this model.
     query = None
 
+    #: arguments for the mapper
+    __mapper_cls__ = _SignalTrackingMapper
+
+    __tablename__ = _ModelTableNameDescriptor()
+
 
 class SQLAlchemy(object):
     """This class is used to control the SQLAlchemy integration to one
@@ -377,11 +477,39 @@ class SQLAlchemy(object):
         class User(db.Model):
             username = db.Column(db.String(80), unique=True)
             pw_hash = db.Column(db.String(80))
+
+    You may also define your own SessionExtension instances as well when
+    defining your SQLAlchemy class instance. You may pass your custom instances
+    to the `session_extensions` keyword. This can be either a single
+    SessionExtension instance, or a list of SessionExtension instances. In the 
+    following use case we use the VersionedListener from the SQLAlchemy
+    versioning examples.::
+
+        from history_meta import VersionedMeta, VersionedListener
+
+        app = Flask(__name__)
+        db = SQLAlchemy(app, session_extensions=[VersionedListener()])
+
+        class User(db.Model):
+            __metaclass__ = VersionedMeta
+            username = db.Column(db.String(80), unique=True)
+            pw_hash = db.Column(db.String(80))
+
+    The `session_options` parameter can be used to override session
+    options.  If provided it's a dict of parameters passed to the
+    session's constructor.
+
+    .. versionadded:: 0.10
+       The `session_options` parameter was added.
     """
 
-    def __init__(self, app=None, use_native_unicode=True):
+    def __init__(self, app=None, use_native_unicode=True,
+                 session_extensions=None, session_options=None):
         self.use_native_unicode = use_native_unicode
-        self.session = _create_scoped_session(self)
+        self.session_extensions = to_list(session_extensions, []) + \
+                                  [_SignallingSessionExtension()]
+
+        self.session = _create_scoped_session(self, session_options)
 
         self.Model = declarative_base(cls=Model, name='Model')
         self.Model.query = _QueryProperty(self)
@@ -434,6 +562,21 @@ class SQLAlchemy(object):
             info.query.setdefault('charset', 'utf8')
             options.setdefault('pool_size', 10)
             options.setdefault('pool_recycle', 7200)
+        elif info.drivername == 'sqlite':
+            pool_size = options.get('pool_size')
+            # we go to memory and the pool size was explicitly set to 0
+            # which is fail.  Let the user know that
+            if info.database in (None, '', ':memory:'):
+                if pool_size == 0:
+                    raise RuntimeError('SQLite in memory database with an '
+                                       'empty queue not possible due to data '
+                                       'loss.')
+            # if pool size is None or explicitly set to 0 we assume the
+            # user did not want a queue for this sqlite connection and
+            # hook in the null pool.
+            elif not pool_size:
+                from sqlalchemy.pool import NullPool
+                options['poolclass'] = NullPool
 
         unu = app.config['SQLALCHEMY_NATIVE_UNICODE']
         if unu is None:
@@ -488,5 +631,5 @@ class SQLAlchemy(object):
                 app = ctx.app
         return '<%s engine=%r>' % (
             self.__class__.__name__,
-            app and self.app.config['SQLALCHEMY_DATABASE_URI'] or None
+            app and app.config['SQLALCHEMY_DATABASE_URI'] or None
         )
